#include <iostream>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <zlib.h>
#include <math.h>
#include <iomanip>
#include <list>
#include <vector>

#include <Inventor/SoInput.h>
#include <Inventor/Xt/SoXt.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/SoInput.h>
#include <Inventor/SoOffscreenRenderer.h>
#include <Inventor/Xt/viewers/SoXtExaminerViewer.h>
#include <Inventor/nodes/SoEventCallback.h>
#include <Inventor/events/SoKeyboardEvent.h>
#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/nodes/SoIndexedFaceSet.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/nodes/SoGroup.h>
#include <Inventor/nodes/SoDirectionalLight.h>
#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/nodes/SoOrthographicCamera.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoResetTransform.h>
#include <Inventor/SoInput.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoSwitch.h>
#include <Inventor/nodes/SoDrawStyle.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoIndexedTriangleStripSet.h>
#include <Inventor/nodes/SoResetTransform.h>
#include <Inventor/nodes/SoCylinder.h>
#include <Inventor/nodes/SoComplexity.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoCone.h>
#include <Inventor/actions/SoWriteAction.h>

using namespace std;

static char version[80] =
"$Revision: 1.4 $";

// variables
bool numSubs = 1;
bool renderQuit = FALSE;
char *outFile = "out.iv";
char *inFile = "scene.iv";
enum subdTypes { CATMULL_CLARK };
enum vertType { ORIGINAL, NEWEDGE, NEWFACE, MOVED };
subdTypes subdType = CATMULL_CLARK;
SoSeparator *scene = NULL;
SoXtExaminerViewer *myViewer;
SoSwitch *wireFrameSwitch;
SoSwitch *shapeSwitch;

struct HE_face;


// clamp definition
// restricts a number to a range [min, max]
template <class T> inline
T clamp(T x, T min, T max){
  if (x<min)
    return(min);
  else if (x>max)
    return(max);
  else
    return(x);
}

typedef struct HE_edge {
  int32_t vert;
  int32_t orig;   // debugging
  HE_edge *pair;
  HE_face *face;
  HE_edge *next;
} HE_edge;

typedef struct HE_face {
  HE_edge *edge;
} HE_face;


//WW''W''''''WWW''''WW'WWWWW'''WW'''''''''''WW''''WW''''''''WW''''''''''WWW''
//WW  W       WW   WW  WW      WW           WW              WW           WW  
//WW  W  WWW  WW  WWWW WWWW   WWW  WWW  WW  WW    WW  WWWW WWWW     WWW  WW  
//WWWWW W WW  WW   WW  WW    W WW W WW WWWW WW    WW WWW    WW     WW W  WW  
//WW  W W WW  WW   WW  WW    W WW W WW WW   WW    WW   WWW  WW     WW    WW  
//WW  W  WWW WWWW  WW  WWWWW  WWW  WWW  WWW WWWWW WW WWWW   WW      WWW WWWW 
//                                  WW                                       
//''''''''''''''''''''''''''''''''WWW''''''''''''''''''''''''''''''''''''''''
//
class HalfEdgeList {

  
private:
  list <HE_edge *> edges;
  list <HE_face *> faces;
  SoIndexedShape *shape;
  SoIndexedFaceSet *faceSet;
  SoMFVec3f *points, *points2;
  vector <vertType> vertTypes;
  
public:

  //WW''''''WW''WW'
  //            WW 
  //WW WWW  WW WWWW
  //WW WW W WW  WW 
  //WW WW W WW  WW 
  //WW WW W WW  WW 
  //               
  //'''''''''''''''
  //
  void init(SoIndexedShape *shape_, SoMFVec3f *points_,
	    SoIndexedFaceSet *faceSet_, SoMFVec3f *points2_){
    shape = shape_;
    points = points_;
    faceSet = faceSet_;
    points2 = points2_;

    // initialize vertex types
    //for (int i=0; i<points2->getNum(); i++)
    //  vertTypes[i] = ORIGINAL;
    vertTypes.clear();
    vertTypes.resize(points2->getNum(), ORIGINAL);
  }


  //WWW'''''''''''''''WW''WWW''WW'''''''''''''''''
  // WW               WW WW  W WW                 
  // WW   WW    WWW  WWW  WW   WWW   WWW WWW   WW 
  // WW  WW W  W WW W WW   WW  WW W W WW WW W WWWW
  // WW  WW W  W WW W WW W  WW WW W W WW WW W WW  
  //WWWW  WW    WWW  WWW  WWW  WW W  WWW WWW   WWW
  //                                     WW       
  //'''''''''''''''''''''''''''''''''''''WW'''''''
  //
  // Loads a single faceset into the data structure.
  // The strategy is to loop over faces, inserting half edges
  // for each edge, which are linked together, including a link to a new face.
  // For convenience, this new face is itself stored in a member list.
  // Additionally, the half edges are stored in lists
  // at each vertex. Then these new half edges are looped over again
  // to compare with the lists at each vertex to identify half-edge pairs.
  // Identified pairs are then removed from the vertex's edge list.
  // In this way each half edge is linked with it's next half edge,
  // a face, a pair, and of course a single vertex.
  //
  int loadShape(){

    int numV = points2->getNum();
    int numI = shape->coordIndex.getNum();
    
    // create half-edge list for each vertex
    list <HE_edge *> vertEdges[numV];

    // state information
    int numPolys = 0;
    int numEdges = 0;
    HE_edge *firstEdge = NULL;
    HE_edge *prevEdge = NULL;

    // indexed face sets
    if (shape->isOfType(SoIndexedFaceSet::getClassTypeId())){

      // loop over polygons
      int j = 0;
      while (j<numI){
      
	if (shape->coordIndex[j] < 0){
	  j++;
	  continue;
	}

	// create new face
	HE_face *face = new HE_face;
	faces.push_back(face);
	numPolys++;
	firstEdge = NULL;

	// since polygon is closed, there is an edge for each vertex
	// loop over vertices, creating edge for each    
	while (1){

	  int v = shape->coordIndex[j];
	
	  // edge found
	  numEdges++;
	  if (numEdges%1000==0)
	    printf("\r%d%    ", numEdges);
	
	  // create half edge
	  HE_edge *edge = new HE_edge;
	  edge->vert = v;
	  //edge->type = ORIGINAL;
	  edge->pair = NULL;
	  edge->face = face;
	
	  // store first edge
	  if (firstEdge == NULL){
	    firstEdge = edge;
	    face->edge = edge;
	  }
	  else {
	    // point last edge to this one
	    prevEdge->next = edge;

	    // debugging
	    edge->orig = prevEdge->vert;
	  }

	  // save ptr to this edge
	  prevEdge = edge;
	
	  j++;

	  // check for end of polygon
	  if (j>=numI || shape->coordIndex[j]<0){

	    // close loop
	    edge->next = firstEdge;               // fill in next

	    // debugging
	    firstEdge->orig = edge->vert;

	    // make second pass to connect pairs
	    // this is done by looping over vertex pairs (found by successive
	    // pairs of half edges), and consuling the 2nd vertex's
	    // edge list for an edge going back to the 1st vertex

	    // loop over edges
	    edge = firstEdge;
	    do {

	      // identify the two vertices
	      int32_t v[2];
	      v[0] = edge->vert;
	      v[1] = edge->next->vert;

	      // search v[1]'s edge list for edge going to v[0]
	      // assign half edge as pair
	      list <HE_edge *>::iterator p;	    
	      for (p=vertEdges[v[1]].begin();
		   p != vertEdges[v[1]].end(); p++){
	      
		HE_edge *edge2 = *p;

		// found edge?
		if (edge2->vert == v[0]){

		  // link edges
		  edge->next->pair = edge2;
		  edge2->pair = edge->next;

		  // remove edge and stop
		  vertEdges[v[1]].erase(p);
		  break;
		}
	      }

	      // if not found, and not paired, add edge to vert's list
	      if (p == vertEdges[v[1]].end() && edge->next->pair == NULL)
		vertEdges[v[0]].push_back(edge->next);

	      // proceed to next edge
	      edge = edge->next;
	    
	    } while (edge != firstEdge);  // end edge loop
	  
	    break; // stop vert looping for this polygon
	  }

	}

      } // end poly loop

    }
    // indexed triangle strip sets
    else if (shape->isOfType(SoIndexedTriangleStripSet::getClassTypeId())){
      
    }
    else {
      cerr << "Unsupported indexed shape: "
	   << shape->getTypeId().getName().getString()
	   << "\n";
    }


    // check for unfound half-edge's
    int unpairedEdges = 0;
    int firstTime = 1;
    for (int i=0; i<numV; i++){

      unpairedEdges += vertEdges[i].size();

      if (unpairedEdges && firstTime){
	cerr << "Unpaired vertices: \n";
	firstTime = 0;
      }

      if (vertEdges[i].size()){
	
	cerr << "Vert " << i << ": ";

	// loop over unapred edges
	for (list <HE_edge *>::iterator p=vertEdges[i].begin();
	     p!=vertEdges[i].end(); ++p){

	  HE_edge *edge = *p;
	  
	  cerr << edge->vert << " ";

	  // debugging
	  show((*points)[edge->orig], (*points)[edge->vert]);
	  
	}
	cerr << "\n";
      }
    }


    //printFaces();
    
    // Euler's formula: F+V-E=2
    // E.g., a cube:    6+8-12=2 
    cerr << numPolys << " faces, "
	 << numV     << " verts, "
         << numEdges << " half edges ("
	 << numEdges/2 << " edges).\n";
    cerr << "Euler characteristic = "
	 << numPolys+numV - numEdges/2 << ".\n";
    cerr << unpairedEdges << " unpaired half-edges.\n";
    cerr << "Done loading face set.\n";

    return(1);
  }


  //''''''''''WW'''''''WW''WWWWW'''''''''''''''''''''
  //                   WW  WW                        
  //WWW  W WW WW WWW  WWWW WWWW   WWW  WWW  WW   WWWW
  //WW W WW   WW WW W  WW  WW    W WW WW W WWWW WWW  
  //WW W WW   WW WW W  WW  WW    W WW WW   WW     WWW
  //WWW  WW   WW WW W  WW  WW     WWW  WWW  WWW WWWW 
  //WW                                               
  //WW'''''''''''''''''''''''''''''''''''''''''''''''
  //
  void printFaces(){

    cerr << "Faces: " << faces.size() << "\n";

    // loop over faces
    for (list <HE_face *>::iterator p=faces.begin(); p != faces.end(); p++){

      // loop over edges
      HE_face *face = *p;
      HE_edge *edge = face->edge;
      cerr << "Face: " << face->edge << "\n";
      do {

	cerr << "Edge: "
	     << edge->vert << " "
	     << edge->face << " "
	     << edge->pair << " "
	     << edge->next << "\n";

	edge = edge->next;
      } while (edge != face->edge);
      
    }
  }


  //'''''''''''WW'''''WW'WW''''''WW'''WW'''''
  //           WW     WW              WW     
  // WWWW W WW WWW   WWW WW WW W WW  WWW  WW 
  //WWW   W WW WW W W WW WW WW W WW W WW WWWW
  //  WWW W WW WW W W WW WW WW W WW W WW WW  
  //WWWW   WWW WWW   WWW WW  WW  WW  WWW  WWW
  //                                         
  //'''''''''''''''''''''''''''''''''''''''''
  //
  void subdivide(){

    // 1. For each face, split edges
    // 2.                create new edges to new face vert
    // 3.                connect sub face edges.
    
    // loop over faces
    list <HE_face *>::iterator firstFace=faces.begin();
    list <HE_face *>::iterator lastFace=faces.end();
    for (list <HE_face *>::iterator p=firstFace; p!=lastFace; p++){

      // loop over edges
      HE_face *face = *p;
      HE_edge *edge = face->edge;

      do {

	// split edge if not already split
	if (vertTypes[edge->vert] != NEWEDGE){

	  // create new edge point
	  int newEdgeVert = points2->getNum();
	  SbVec3f newEdgePos = ((*points2)[edge->vert]+(*points2)[edge->pair->vert])/2.0;
	  points2->set1Value(newEdgeVert, newEdgePos);
	  //vertTypes[newEdgeVert] = NEWEDGE;
	  vertTypes.push_back(NEWEDGE);

	  //
	  //     -------->0    pair
	  //    0<--------     edge
	  //
	  //         |
	  //         V
	  // 
	  //     old  new
	  //     --->O--->O    pair
	  //    0<---O<---     edge
	  //      new  old
	  //
	  
	  // point new edge at current edge vert, and current edge at new vert
	  HE_edge *newEdge = new HE_edge;
	  //newEdge->type = ORIGINAL;
	  newEdge->vert = edge->vert;
	  newEdge->face = face;
	  newEdge->pair = edge->pair;
	  newEdge->next = edge->next;

	  HE_edge *newEdgePair = new HE_edge;
	  //newEdgePair->type = ORIGINAL;
	  newEdgePair->vert = edge->pair->vert;
	  newEdgePair->face = edge->pair->face;
	  newEdgePair->pair = edge;
	  newEdgePair->next = edge->pair->next;

	  //edge->pair->type = NEWEDGE;
	  edge->pair->vert = newEdgeVert;
	  edge->pair->next = newEdgePair;
	  edge->pair->pair = newEdge;
	  
	  //edge->type = NEWEDGE;
	  edge->vert = newEdgeVert;
	  edge->next = newEdge;
	  edge->pair = newEdgePair;

	}

	// advance edge pointer to next side
	edge = edge->next->next;
	
      } while (edge != face->edge);

      
      // create new face point
      int newFaceVert = points2->getNum();
      int numFaceVerts = 0;
      SbVec3f newFacePos(0.0, 0.0, 0.0);
      points2->set1Value(newFaceVert, newFacePos);
      //vertTypes[newFaceVert] = NEWFACE;
      vertTypes.push_back(NEWFACE);
      
      // loop around face again, creating new faces
      edge = face->edge;
      do {

	HE_edge *nextEdge = edge->next;
	
	if (vertTypes[edge->vert] == NEWEDGE){

	  HE_face *newFace = new HE_face;

	  newFacePos += (*points2)[edge->next->vert];
	  numFaceVerts++;

          //       edge
	  //  O<--O<--O
	  //  |out^|in^
	  //  v   |v  |
	  //  O   ()  O
	  //  |       ^
	  //  v       |
	  //  O-->O-->O
	  
	  
	  // create two new edges
	  HE_edge *newEdgeIn = new HE_edge;
	  HE_edge *newEdgeOut = new HE_edge;
	  
	  //newEdgeIn->type = NEWFACE;
	  newEdgeIn->vert = newFaceVert;
	  newEdgeIn->face = edge->next->face;
	  newEdgeIn->pair = newEdgeOut;
	  newEdgeIn->next = NULL;    // filled-in in next pass

	  //newEdgeOut->type = NEWEDGE;
	  newEdgeOut->vert = edge->vert;
	  newEdgeOut->face = newFace;
	  newEdgeOut->pair = newEdgeIn;
	  newEdgeOut->next = edge->next;

	  // configure face
	  newFace->edge = newEdgeOut;
	  faces.push_front(newFace);

	  newEdgeOut->next->face = newFace;
	  newEdgeOut->next->next->face = newFace;
	  newEdgeOut->next->next->next->face = newFace;

	  // splice new edges into edge/edge->next
	  edge->next = newEdgeIn;

	}
	
	edge = nextEdge;
      } while (edge != face->edge);

      // loop around face again, filling-in newEdgeIn's next field
      edge = face->edge;
      do {

	edge->next->pair->next->next->next->next = edge->next->pair;
	
	edge = edge->next->pair->next->next;
      } while (edge != face->edge);

      // remove old face
      delete face;

      // set face position
      newFacePos /= numFaceVerts;
      points2->set1Value(newFaceVert, newFacePos);
    }

    // remove faces
    faces.erase(firstFace, lastFace);
    
  }


  //'''''''''''WW'''''''''''''''WWW''WWW'''WWW''WWW''''''''''''WW'''
  //           WW                WW   WW  WW  W  WW            WW   
  // WWW  WWW WWWW WWWW W  W WW  WW   WW  WW     WW   WWW W WW WW W 
  //WW W W WW  WW  WW WW W W WW  WW   WW  WW     WW  W WW WW   WWW  
  //WW   W WW  WW  WW WW W W WW  WW   WW  WW  W  WW  W WW WW   WW W 
  // WWW  WWW  WW  WW WW W  WWW WWWW WWWW  WWW  WWWW  WWW WW   WW  W
  //                                                                
  //''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
  //
  void catmullClark(){

    // copy points2
    SoMFVec3f *points3 = new SoMFVec3f();
    points3->setValues(0, points2->getNum(), points2->getValues(0));    
    
    // loop over faces
    for (list <HE_face *>::iterator p=faces.begin(); p!=faces.end(); p++){

      // loop over edges
      HE_face *face = *p;
      HE_edge *edge = face->edge;

      do {

	if (vertTypes[edge->vert] == ORIGINAL){

	  // calculate edge and face averages
	  SbVec3f faceAvgPos(0.0, 0.0, 0.0);
	  SbVec3f edgeAvgPos(0.0, 0.0, 0.0);
	  int edges = 0;

	  // loop over edges around original vert
	  HE_edge *newEdge = edge->pair;
	  do {

	    if (vertTypes[newEdge->vert] == NEWEDGE){
	      SbVec3f newPos =
		(*points3)[newEdge->pair->vert] +
		(*points3)[newEdge->next->vert] +
		(*points3)[newEdge->next->pair->next->vert] +
		(*points3)[newEdge->next->pair->next->pair->next->vert];
	      newPos /= 4.0;
	      
	      points2->set1Value(newEdge->vert, newPos);	      

	      // mark as moved in all half edges pointing to it
	      vertTypes[newEdge->vert] = MOVED;
	    }


	    // track edge position average
	    edgeAvgPos += (*points3)[newEdge->vert];
	    
	    // track face position average
	    faceAvgPos += (*points3)[newEdge->next->vert];

	    edges++;
	    
	    newEdge = newEdge->pair->next;
	  } while (newEdge != edge->pair);

	  edgeAvgPos /= edges;
	  faceAvgPos /= edges;
	  
	  // move original vert
	  SbVec3f newOrigPos = (faceAvgPos + 2.0*edgeAvgPos +
				(edges-3)*(*points2)[edge->vert])/edges;
	  points2->set1Value(edge->vert, newOrigPos);
	  vertTypes[edge->vert] = MOVED;
	}

	edge = edge->next;
      } while (edge != face->edge);
    
    }
    

    // free memory
    delete points3;
    
  }

  SoIndexedFaceSet* getFaceSet(){

    // read half edges into face set
    faceSet->coordIndex.setNum(0);

    // loop over faces
    int i = 0;
    for (list <HE_face *>::iterator p=faces.begin(); p != faces.end(); p++){

      // loop over edges
      HE_edge *edge = (*p)->edge;
      do {

	faceSet->coordIndex.set1Value(i++, edge->vert);
	vertTypes[edge->vert] = ORIGINAL;
	edge = edge->next;
	
      } while (edge != (*p)->edge);

      faceSet->coordIndex.set1Value(i++, -1);
    }
    
    return(faceSet);
  }

  // display edge
  void show(SbVec3f v0, SbVec3f v1){

    int clear = 0;
    SbColor color = SbColor(1.0, 0.0, 0.0);
    SbVec3f dir = v1-v0;
    float height = dir.normalize();
    
    static double hue = 0.0;
    double red, green, blue;

    // height stuff
    double cb_rat = 0.05;
    double cone_height = cb_rat*height;
    double bar_height = height-cone_height;
    double bar_radius = bar_height/100.0;
    double cone_radius = bar_height/30.0;

    // add self to scene graph
    //pthread_mutex_lock(&lock_show);
    static SoSeparator *showRaySep = NULL;
    if (showRaySep == NULL){
      // add cube sep to scene
      showRaySep = new SoSeparator();
      showRaySep->addChild(new SoResetTransform());
      SoComplexity *comp = new SoComplexity();
      comp->value.setValue(0.3);
      showRaySep->addChild(comp);
      scene->addChild(showRaySep);
    }
    else if (clear){
      showRaySep->removeAllChildren();
      SoComplexity *comp = new SoComplexity();
      comp->value.setValue(0.3);
      showRaySep->addChild(comp);

      //pthread_mutex_unlock(&lock_show);
      return;
    }
    //pthread_mutex_unlock(&lock_show);
  
    // Ray debugging
    SoSeparator *sep = new SoSeparator();
    SoTransform *trn;
    SoMaterial *mat = new SoMaterial();
    mat->ambientColor.setValue(0.0, 0.0, 0.0);
    mat->diffuseColor.setValue(0.0, 0.0, 0.0);
    mat->specularColor.setValue(1.0, 1.0, 1.0);
    mat->shininess.setValue(0.3);
    mat->transparency.setValue(0.5);

    if (color[0]<0.0){
      //printCol(color);

      //pthread_mutex_lock(&lock_show);
      hue += (1.0-hue)*0.05;
      mat->emissiveColor.setHSVValue(hue, 1.0, 0.8);
      //pthread_mutex_unlock(&lock_show);
    }
    else {
      for (int i=0; i<3; i++)
//	color[i] = 1.0-exp(-2.0*color[i]/10.0);
      color[i] = clamp(color[i], 0.0f, 1.0f);
      //mat->diffuseColor.setValue(color);
      mat->emissiveColor.setValue(color);
    }
    sep->addChild(mat);
    
    // Translate base
    trn = new SoTransform();
    trn->translation.setValue(v0);
    sep->addChild(trn);

  
    // Scale
    trn = new SoTransform();

    // Rotate
    SbRotation sbr;
    sbr.setValue(SbVec3f(0.0, 1.0, 0.0), dir);
    trn->rotation.setValue(sbr);
    sep->addChild(trn);

    // Pole
    SoCylinder *cyl = new SoCylinder();
    cyl->radius.setValue(bar_radius);
    cyl->height.setValue(bar_height);
    trn = new SoTransform();
    trn->translation.setValue(SbVec3f(0.0, bar_height/2.0, 0.0));
    sep->addChild(trn);
    sep->addChild(cyl);
  
    // Cone
    SoCone *con = new SoCone();
    con->bottomRadius.setValue(cone_radius);
    con->height.setValue(cone_height);
    trn = new SoTransform();
    trn->translation.setValue(SbVec3f(0.0, height/2.0, 0.0));
    sep->addChild(trn);
    sep->addChild(con);
  
    //pthread_mutex_lock(&lock_show);
    showRaySep->addChild(sep);
    //pthread_mutex_unlock(&lock_show);
  }

  
};

HalfEdgeList HE_list;


//WW''''''''''''''WWWWW'''''''''''''''''WW''WW''W'''''''''''''WW'WWW'''''''
//WW              WW                    WW  WW  W             WW  WW       
//WW W   WW  W WW WWWW  WW W  WW  WWW  WWWW WW  W  WWW WWW   WWW  WW   WW  
//WWW   WWWW W WW WW    WW W WWWW WW W  WW  WWWWW W WW WW W W WW  WW  WWWW 
//WW W  WW   W WW WW    WW W WW   WW W  WW  WW  W W WW WW W W WW  WW  WW   
//WW  W  WWW  WWW WWWWW  WW   WWW WW W  WW  WW  W  WWW WW W  WWW WWWW  WWW 
//             WW                                                          
//'''''''''''WWW'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
//
void keyEventHandler(void *gvp, SoEventCallback *callBack){
  const SoEvent *event = callBack->getEvent();

  // hide wireframe
  if (SO_KEY_PRESS_EVENT(event, H)){
    if (shapeSwitch->whichChild.getValue() == 0){
      cerr << "Hiding shape.\n";
      shapeSwitch->whichChild.setValue(-1);
    }
    else {
      cerr << "Showing shape.\n";
      shapeSwitch->whichChild.setValue(0);
    }
  }
  
  // subdivide
  if (SO_KEY_PRESS_EVENT(event, S)){
    cerr << "Subdividing.\n";
    wireFrameSwitch->whichChild.setValue(0);
    HE_list.subdivide();
    //HE_list.printFaces();
    HE_list.getFaceSet();
  }

  // refine
  if (SO_KEY_PRESS_EVENT(event, R)){
    cerr << "Refining.\n";
    wireFrameSwitch->whichChild.setValue(0);
    HE_list.subdivide();
    HE_list.catmullClark();
    HE_list.getFaceSet();
  }

  // Save scene to file
  if (SO_KEY_PRESS_EVENT(event, W)){

    SoOutput *out = new SoOutput();
    out->openFile(outFile);
    SoWriteAction *wa = new SoWriteAction(out);
    wa->apply(scene);
    out->closeFile();
    cerr << "Dumped scene graph to " << outFile << ".\n";
    delete out;
    delete wa;
  }
  
}

//'''''''''''''WW'''''
//                    
//WWWW W   WWW WW WWW 
//WW WW W W WW WW WW W
//WW WW W W WW WW WW W
//WW WW W  WWW WW WW W
//                    
//''''''''''''''''''''
//
int main(int argc, char *argv[]){

  // read options
  while (1){
    char c = getopt(argc, argv, "c:hld:s:o:r");

    if (c == -1)
      break;

    switch (c){

    case 'h':
       cerr << "Usage: " << argv[0]
	 << " [-h] [-o out.iv] [-nN] [-r] [-t] [scene.iv]\n"
	 << "\n"
	 << "Loads .iv scene file and subdivides it.\n"
         << "Quads in SoIndexedShapes only! for right now.\n"
	 << ""
	 << "\n"
	 << " -h              this help message\n"
	 << " -o file         name of output rgb file (out.iv)\n"
	 << " -nN             use N subdivisions (1)\n"
	 << " -r              render straight to file without viewer.\n"
	 << " -t TYPE         use TYPE subdivision method (catmullClark)\n";
	 
      exit(1);
      break;

    case 'o':
      outFile = optarg;
      break;
      
    case 'n':
      numSubs = atoi(optarg);
      break;

     case 'r':
       renderQuit = TRUE;
       break;

    case 't':
      if (strcmp(optarg, "catmullClark"))
	subdType = CATMULL_CLARK;
      break;
       
     default:
       break;
    }
  }

  // read input filename
  if (optind<argc){
    inFile = argv[optind++];
  }
  
  while (optind<argc){
    cerr << "Unknown option: " << argv[optind++] << "\n";
  }
  
  // initialize inventor
  if (renderQuit)
    SoDB::init();
  else
    SoXt::init(argv[0]);

  // open scene
  SoInput *sceneIn = new SoInput();
  if (!sceneIn->openFile(inFile)){
    cerr << "Unable to open scene file: " << inFile << endl;
    exit(1);
  }

  // read scene
  scene = SoDB::readAll(sceneIn);
  if (!scene){
    cerr << "Error: scene separator is null!" << endl;
    exit(1);
  }
  scene->ref();   // ref count is initially zero (why I dunno...)
  delete sceneIn;

  // Search for indexed shapes in scene graph
  SoSearchAction sa;
  sa.setType(SoIndexedShape::getClassTypeId());
  sa.setInterest(SoSearchAction::ALL);
  sa.apply(scene);
  SoPathList plist = sa.getPaths();
  int len = plist.getLength();

  if (len<=0){
    cerr << "Could not find SoIndexedShape in " << inFile << "\n";
    exit(1);
  }


  // find face set
  SoPath *path = (SoPath*) plist[0];
  SoIndexedShape *shape = (SoIndexedShape *)path->getTail();
  SoGroup *shapeParent = (SoGroup *)path->getNodeFromTail(1);

  // put in group
  shapeSwitch = new SoSwitch();
  shapeSwitch->addChild(shape);
  shapeSwitch->whichChild.setValue(0);
  shapeParent->replaceChild(shape, shapeSwitch);
  
  // insert wireframe style and base light model
  wireFrameSwitch = new SoSwitch();
  SoGroup *wireFrameGroup = new SoGroup();
  SoLightModel *baseModel = new SoLightModel();
  baseModel->model.setValue(SoLightModel::BASE_COLOR);
  wireFrameGroup->addChild(baseModel);
  SoDrawStyle *wireFrameStyle = new SoDrawStyle();
  wireFrameStyle->style.setValue(SoDrawStyle::LINES);
  wireFrameGroup->addChild(wireFrameStyle);
  wireFrameSwitch->addChild(wireFrameGroup);
  int shapeIdx = shapeParent->findChild(shapeSwitch);
  shapeParent->insertChild(wireFrameSwitch, shapeIdx++);

  // insert extra light model and draw style
  shapeParent->insertChild(new SoLightModel(), shapeIdx+1);
  shapeParent->insertChild(new SoDrawStyle(), shapeIdx+2);

  
  // find vertex list
  SoMFVec3f *points = NULL;
  
  // check for vertex property
  if (shape->vertexProperty.getValue() == NULL){

    // search for Coordinate3 node
    cerr << "Searching for Coordinate3 node...";
    SoSearchAction sa2;
    sa2.setType(SoCoordinate3::getClassTypeId());
    sa2.setSearchingAll(FALSE); // don't look under off switches
    sa2.setInterest(SoSearchAction::LAST);
    sa2.apply(path);

    SoPath *coordPath = sa2.getPath();
    if (coordPath == NULL){
      cerr << "Could not find Coordinate3 node!\n";
      exit(2);
    }
    else
      cerr << "found.\n";
      
    SoCoordinate3 *coord = (SoCoordinate3 *)coordPath->getTail();
    points = &(coord->point);
    
  }
  else {
    SoVertexProperty *vProp = (SoVertexProperty *)shape->vertexProperty.getValue();
    points = &(vProp->vertex);
  }

  // insert secondary face set
  SoIndexedFaceSet *faceSet = new SoIndexedFaceSet();
  SoVertexProperty *vProp2 = new SoVertexProperty();
  SoMFVec3f *points2 = &(vProp2->vertex);
  points2->setValues(0, points->getNum(), points->getValues(0));
  faceSet->vertexProperty.setValue(vProp2);
  shapeParent->insertChild(faceSet, shapeIdx+3);
  
  // load into half-edge data structure
  HE_list.init(shape, points, faceSet, points2);
  HE_list.loadShape();

  if (!renderQuit){
    // Set up keyboard
    SoEventCallback *callBack = new SoEventCallback();
    callBack->addEventCallback(SoKeyboardEvent::getClassTypeId(),
			       keyEventHandler);
    scene->addChild(callBack);
    cerr << "\nPress ESC, then R to save to file.\n" << endl;

    // This returns a main window to use.
    // If unsuccessful, exit.
    Widget myWindow = SoXt::init(argv[0]); // pass the app name
    if (myWindow == NULL) exit(1);

    // Set up viewer
    myViewer = new SoXtExaminerViewer(myWindow);
    myViewer->setSceneGraph(scene);
    myViewer->setTitle("Subdivider");
    myViewer->show();
    
    SoXt::show(myWindow); // Display main window
    SoXt::mainLoop();	// Main Inventor event loop
  }
  else
    1;
  
  return(0);
}
